@@ -1,14 +1,14 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2007-2019 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2007-2021 -- leonerd@leonerd.org.uk
 
 package IO::Async::Loop;
 
 use strict;
 use warnings;
 
-our $VERSION = '0.75';
+our $VERSION = '0.802';
 
 # When editing this value don't forget to update the docs below
 use constant NEED_API_VERSION => '0.33';
@@ -23,6 +23,9 @@ use constant _CAN_SUBSECOND_ACCURATELY => 0;
 # Does the loop implementation support IO_ASYNC_WATCHDOG?
 use constant _CAN_WATCHDOG => 0;
 
+# Does the loop support ->watch_process on PID 0 to observe all exits?
+use constant _CAN_WATCH_ALL_PIDS => 1;
+
 # Watchdog configuration constants
 use constant WATCHDOG_ENABLE   => $ENV{IO_ASYNC_WATCHDOG};
 use constant WATCHDOG_INTERVAL => $ENV{IO_ASYNC_WATCHDOG_INTERVAL} || 10;
@@ -36,6 +39,7 @@ use Scalar::Util qw( refaddr weaken );
 use Socket qw( SO_REUSEADDR AF_INET6 IPPROTO_IPV6 IPV6_V6ONLY );
 
 use IO::Async::OS;
+use IO::Async::Metrics '$METRICS';
 
 use constant HAVE_SIGNALS => IO::Async::OS->HAVE_SIGNALS;
 use constant HAVE_POSIX_FORK => IO::Async::OS->HAVE_POSIX_FORK;
@@ -86,31 +90,31 @@ C<IO::Async::Loop> - core loop of the C<IO::Async> framework
 
 =head1 SYNOPSIS
 
- use IO::Async::Stream;
- use IO::Async::Timer::Countdown;
+   use IO::Async::Stream;
+   use IO::Async::Timer::Countdown;
 
- use IO::Async::Loop;
+   use IO::Async::Loop;
 
- my $loop = IO::Async::Loop->new;
+   my $loop = IO::Async::Loop->new;
 
- $loop->add( IO::Async::Timer::Countdown->new(
-    delay => 10,
-    on_expire => sub { print "10 seconds have passed\n" },
- )->start );
+   $loop->add( IO::Async::Timer::Countdown->new(
+      delay => 10,
+      on_expire => sub { print "10 seconds have passed\n" },
+   )->start );
 
- $loop->add( IO::Async::Stream->new_for_stdin(
-    on_read => sub {
-       my ( $self, $buffref, $eof ) = @_;
+   $loop->add( IO::Async::Stream->new_for_stdin(
+      on_read => sub {
+         my ( $self, $buffref, $eof ) = @_;
 
-       while( $$buffref =~ s/^(.*)\n// ) {
-          print "You typed a line $1\n";
-       }
+         while( $$buffref =~ s/^(.*)\n// ) {
+            print "You typed a line $1\n";
+         }
 
-       return 0;
-    },
- ) );
+         return 0;
+      },
+   ) );
 
- $loop->run;
+   $loop->run;
 
 =head1 DESCRIPTION
 
@@ -173,6 +177,8 @@ sub __new
       os            => {}, # A generic scratchpad for IO::Async::OS to store whatever it wants
    }, $class;
 
+   $METRICS and $METRICS->inc_gauge( loops => [ class => ref $self ] );
+
    # It's possible this is a specific subclass constructor. We still want the
    # magic IO::Async::Loop->new constructor to yield this if it's the first
    # one
@@ -193,6 +199,13 @@ sub __new
    return $self;
 }
 
+sub DESTROY
+{
+   my $self = shift;
+
+   $METRICS and $METRICS->dec_gauge( loops => [ class => ref $self ] );
+}
+
 =head1 MAGIC CONSTRUCTOR
 
 =head2 new
@@ -214,12 +227,12 @@ module for another event system.
 For example, the following two C<$loop> variables will refer to the same
 object:
 
- use IO::Async::Loop;
- use IO::Async::Loop::Poll;
+   use IO::Async::Loop;
+   use IO::Async::Loop::Poll;
 
- my $loop_poll = IO::Async::Loop::Poll->new;
+   my $loop_poll = IO::Async::Loop::Poll->new;
 
- my $loop = IO::Async::Loop->new;
+   my $loop = IO::Async::Loop->new;
 
 While it is not advised to do so under normal circumstances, if the program
 really wishes to construct more than one Loop object, it can call the
@@ -402,6 +415,7 @@ sub _add_noparentcheck
    my $nkey = refaddr $notifier;
 
    $self->{notifiers}->{$nkey} = $notifier;
+   $METRICS and $METRICS->inc_gauge( notifiers => );
 
    $notifier->__set_loop( $self );
 
@@ -441,6 +455,7 @@ sub _remove_noparentcheck
    exists $self->{notifiers}->{$nkey} or croak "Notifier does not exist in collection";
 
    delete $self->{notifiers}->{$nkey};
+   $METRICS and $METRICS->dec_gauge( notifiers => );
 
    $notifier->__set_loop( undef );
 
@@ -630,7 +645,7 @@ sub new_future
 Blocks until the given future is ready, as indicated by its C<is_ready> method.
 As a convenience it returns the future, to simplify code:
 
- my @result = $loop->await( $future )->get;
+   my @result = $loop->await( $future )->get;
 
 =cut
 
@@ -774,7 +789,7 @@ A CODE reference to the handling callback.
 =back
 
 Attaching to C<SIGCHLD> is not recommended because of the way all child
-processes use it to report their termination. Instead, the C<watch_child>
+processes use it to report their termination. Instead, the C<watch_process>
 method should be used to watch for termination of a given child process. A
 warning will be printed if C<SIGCHLD> is passed here, but in future versions
 of L<IO::Async> this behaviour may be disallowed altogether.
@@ -794,9 +809,9 @@ sub attach_signal
    HAVE_SIGNALS or croak "This OS cannot ->attach_signal";
 
    if( $signal eq "CHLD" ) {
-      # We make special exception to allow $self->watch_child to do this
+      # We make special exception to allow $self->watch_process to do this
       caller eq "IO::Async::Loop" or
-         carp "Attaching to SIGCHLD is not advised - use ->watch_child instead";
+         carp "Attaching to SIGCHLD is not advised - use ->watch_process instead";
    }
 
    if( not $self->{sigattaches}->{$signal} ) {
@@ -861,6 +876,8 @@ sub detach_signal
 
    $loop->later( $code )
 
+   $f = $loop->later
+
 Schedules a code reference to be invoked as soon as the current round of IO
 operations is complete.
 
@@ -874,6 +891,11 @@ This method is implemented using the C<watch_idle> method, with the C<when>
 parameter set to C<later>. It will return an ID value that can be passed to
 C<unwatch_idle> if required.
 
+I<Since version 0.78>: If no C<$code> value is passed, a L<Future> will be
+returned instead. This allows for constructs such as:
+
+   await $loop->later;
+
 =cut
 
 sub later
@@ -881,7 +903,17 @@ sub later
    my $self = shift;
    my ( $code ) = @_;
 
-   return $self->watch_idle( when => 'later', code => $code );
+   return $self->watch_idle( when => 'later', code => $code )
+      if $code;
+
+   my $f = $self->new_future;
+   my $id = $self->watch_idle( when => 'later', code => sub {
+      $f->done unless $f->is_ready;
+   } );
+   $f->on_cancel( sub {
+      $self->unwatch_idle( $id );
+   } );
+   return $f;
 }
 
 =head2 spawn_child
@@ -914,7 +946,7 @@ process before running the code or command. See below.
 A continuation to be called when the child processes exits. It will be invoked
 in the following way:
 
- $on_exit->( $pid, $exitcode, $dollarbang, $dollarat )
+   $on_exit->( $pid, $exitcode, $dollarbang, $dollarat )
 
 The second argument is passed the plain perl C<$?> value.
 
@@ -987,7 +1019,7 @@ Shortcut for passing C<fdI<n>>, where I<n> is the fileno of the IO
 reference. In this case, the key must be a reference that implements the
 C<fileno> method. This is mostly useful for
 
- $handle => 'keep'
+   $handle => 'keep'
 
 =item fdI<n> => IO
 
@@ -1010,10 +1042,10 @@ existing one. If you want to simply add new keys or change the values of some
 keys without removing the other existing ones, you can simply copy C<%ENV>
 into the hash before setting new keys:
 
- env => {
-    %ENV,
-    ANOTHER => "key here",
- }
+   env => {
+      %ENV,
+      ANOTHER => "key here",
+   }
 
 =item nice => INT
 
@@ -1340,7 +1372,7 @@ C<run_process>:
 A continuation to be called when the child process exits and closed its STDOUT
 and STDERR streams. It will be invoked in the following way:
 
- $on_finish->( $pid, $exitcode, $stdout, $stderr )
+   $on_finish->( $pid, $exitcode, $stdout, $stderr )
 
 The second argument is passed the plain perl C<$?> value.
 
@@ -1486,26 +1518,33 @@ format recognised by L<IO::Async::OS>'s C<extract_addrinfo> method.
 This example shows how to use the C<Socket> functions to construct one for TCP
 port 8001 on address 10.0.0.1:
 
- $loop->connect(
-    addr => {
-       family   => "inet",
-       socktype => "stream",
-       port     => 8001,
-       ip       => "10.0.0.1",
-    },
-    ...
- );
+   $loop->connect(
+      addr => {
+         family   => "inet",
+         socktype => "stream",
+         port     => 8001,
+         ip       => "10.0.0.1",
+      },
+      ...
+   );
 
 This example shows another way to connect to a UNIX socket at F<echo.sock>.
 
- $loop->connect(
-    addr => {
-       family   => "unix",
-       socktype => "stream",
-       path     => "echo.sock",
-    },
-    ...
- );
+   $loop->connect(
+      addr => {
+         family   => "unix",
+         socktype => "stream",
+         path     => "echo.sock",
+      },
+      ...
+   );
+
+=item peer => IO
+
+Shortcut for constructing an address to connect to the given IO handle, which
+must be a L<IO::Socket> or subclass, and is presumed to be a local listening
+socket (perhaps on C<PF_UNIX> or C<PF_INET>). This is convenient for
+connecting to a local filehandle, for example during a unit test or similar.
 
 =item local_addrs => ARRAY
 
@@ -1580,11 +1619,11 @@ this callback is invoked to inform of the error. It is passed the name of the
 syscall that failed, the arguments that were passed to it, and the error it
 generated. I.e.
 
- $on_fail->( "socket", $family, $socktype, $protocol, $! );
+   $on_fail->( "socket", $family, $socktype, $protocol, $! );
 
- $on_fail->( "bind", $sock, $address, $! );
+   $on_fail->( "bind", $sock, $address, $! );
 
- $on_fail->( "connect", $sock, $address, $! );
+   $on_fail->( "connect", $sock, $address, $! );
 
 Because of the "try all" nature when given a list of multiple addresses, this
 callback may be invoked multiple times, even before an eventual success.
@@ -1609,7 +1648,7 @@ A continuation that is invoked on a successful C<connect(2)> call to a valid
 socket. It will be passed the connected socket handle, as an C<IO::Socket>
 object.
 
- $on_connected->( $handle )
+   $on_connected->( $handle )
 
 =item on_stream => CODE
 
@@ -1618,14 +1657,14 @@ of L<IO::Async::Stream> when the socket is connected. This is provided as a
 convenience for the common case that a Stream object is required as the
 transport for a Protocol object.
 
- $on_stream->( $stream )
+   $on_stream->( $stream )
 
 =item on_socket => CODE
 
 Similar to C<on_stream>, but constructs an instance of L<IO::Async::Socket>.
 This is most useful for C<SOCK_DGRAM> or C<SOCK_RAW> sockets.
 
- $on_socket->( $socket )
+   $on_socket->( $socket )
 
 =item on_connect_error => CODE
 
@@ -1635,7 +1674,7 @@ occurred, and the name of the operation it occurred in. Errors from the
 C<connect(2)> syscall are considered most significant, then C<bind(2)>, then
 finally C<socket(2)>.
 
- $on_connect_error->( $syscall, $! )
+   $on_connect_error->( $syscall, $! )
 
 =item on_resolve_error => CODE
 
@@ -1826,13 +1865,13 @@ Optional. A callback that is invoked if a syscall fails while attempting to
 create a listening sockets. It is passed the name of the syscall that failed,
 the arguments that were passed to it, and the error generated. I.e.
 
- $on_fail->( "socket", $family, $socktype, $protocol, $! );
+   $on_fail->( "socket", $family, $socktype, $protocol, $! );
 
- $on_fail->( "sockopt", $sock, $optname, $optval, $! );
+   $on_fail->( "sockopt", $sock, $optname, $optval, $! );
 
- $on_fail->( "bind", $sock, $address, $! );
+   $on_fail->( "bind", $sock, $address, $! );
 
- $on_fail->( "listen", $sock, $queuesize, $! );
+   $on_fail->( "listen", $sock, $queuesize, $! );
 
 =item queuesize => INT
 
@@ -1873,7 +1912,7 @@ continuations to invoke on success or failure.
 Optional. A callback that is invoked when the Listener object is ready to
 receive connections. The callback is passed the Listener object itself.
 
- $on_notifier->( $listener )
+   $on_notifier->( $listener )
 
 If this callback is required, it may instead be better to construct the
 Listener object directly.
@@ -1884,7 +1923,7 @@ Optional. A callback that is invoked when the listening socket is ready.
 Typically this would be used in the name resolver case, in order to inspect
 the socket's sockname address, or otherwise inspect the filehandle.
 
- $on_listen->( $socket )
+   $on_listen->( $socket )
 
 =item on_listen_error => CODE
 
@@ -2163,12 +2202,12 @@ thows an exception).
 A optional continuation to be called when the child processes exits. It will
 be invoked in the following way:
 
- $on_exit->( $pid, $exitcode )
+   $on_exit->( $pid, $exitcode )
 
 The second argument is passed the plain perl C<$?> value.
 
 This key is optional; if not supplied, the calling code should install a
-handler using the C<watch_child> method.
+handler using the C<watch_process> method.
 
 =item keep_signals => BOOL
 
@@ -2212,9 +2251,11 @@ sub fork
    }
 
    if( defined $params{on_exit} ) {
-      $self->watch_child( $kid => $params{on_exit} );
+      $self->watch_process( $kid => $params{on_exit} );
    }
 
+   $METRICS and $METRICS->inc_counter( forks => );
+
    return $kid;
 }
 
@@ -2244,11 +2285,11 @@ C<scalar> if not supplied.
 Callback to invoke when the thread function returns or throws an exception.
 If it returned, this callback will be invoked with its result
 
- $on_joined->( return => @result )
+   $on_joined->( return => @result )
 
 If it threw an exception the callback is invoked with the value of C<$@>
 
- $on_joined->( died => $! )
+   $on_joined->( died => $! )
 
 =back
 
@@ -2356,10 +2397,23 @@ The current API version is C<0.49>.
 
 This method may be implemented using C<constant>; e.g
 
- use constant API_VERSION => '0.49';
+   use constant API_VERSION => '0.49';
 
 =cut
 
+sub pre_wait
+{
+   my $self = shift;
+   $METRICS and $self->{processing_start} and
+      $METRICS->report_timer( processing_time => Time::HiRes::tv_interval $self->{processing_start} );
+}
+
+sub post_wait
+{
+   my $self = shift;
+   $METRICS and $self->{processing_start} = [ Time::HiRes::gettimeofday ];
+}
+
 =head2 watch_io
 
    $loop->watch_io( %params )
@@ -2877,9 +2931,9 @@ sub _reap_children
    }
 }
 
-=head2 watch_child
+=head2 watch_process
 
-   $loop->watch_child( $pid, $code )
+   $loop->watch_process( $pid, $code )
 
 This method adds a new handler for the termination of the given child process
 PID, or all child processes.
@@ -2894,7 +2948,7 @@ The PID to watch. Will report on all child processes if this is 0.
 
 A CODE reference to the exit handler. It will be invoked as
 
- $code->( $pid, $? )
+   $code->( $pid, $? )
 
 The second argument is passed the plain perl C<$?> value.
 
@@ -2902,9 +2956,9 @@ The second argument is passed the plain perl C<$?> value.
 
 After invocation, the handler for a PID-specific watch is automatically
 removed. The all-child watch will remain until it is removed by
-C<unwatch_child>.
+C<unwatch_process>.
 
-This and C<unwatch_child> are optional; a subclass may implement neither, or
+This and C<unwatch_process> are optional; a subclass may implement neither, or
 both. If it implements neither then child watching will be performed by using
 C<watch_signal> to install a C<SIGCHLD> handler, which will use C<waitpid> to
 look for exited child processes.
@@ -2912,13 +2966,26 @@ look for exited child processes.
 If both a PID-specific and an all-process watch are installed, there is no
 ordering guarantee as to which will be called first.
 
+B<NOTE> that not all loop classes may be able to support the all-child watch.
+The basic Select and Poll-based classes provided by this distribution do, and
+those built on top of similar OS-specific mechanisms such as Linux's Epoll
+probably will, but typically those built on top of other event systems such
+as F<glib> or F<libuv> may not be able, as the underlying event system may not
+provide the necessary hooks to support it.
+
 =cut
 
-sub watch_child
+sub watch_process
 {
    my $self = shift;
    my ( $pid, $code ) = @_;
 
+   if( $self->API_VERSION < 0.76 and
+      ( $self->can( "watch_child" ) // 0 ) != \&watch_child ) {
+      # Invoke legacy loop API
+      return $self->watch_child( @_ );
+   }
+
    my $childwatches = $self->{childwatches};
 
    croak "Already have a handler for $pid" if exists $childwatches->{$pid};
@@ -2940,19 +3007,28 @@ sub watch_child
    $childwatches->{$pid} = $code;
 }
 
-=head2 unwatch_child
+# Old name
+sub watch_child { shift->watch_process( @_ ) }
 
-   $loop->unwatch_child( $pid )
+=head2 unwatch_process
+
+   $loop->unwatch_process( $pid )
 
 This method removes a watch on an existing child process PID.
 
 =cut
 
-sub unwatch_child
+sub unwatch_process
 {
    my $self = shift;
    my ( $pid ) = @_;
 
+   if( $self->API_VERSION < 0.76 and
+      ( $self->can( "unwatch_child" ) // 0 ) != \&unwatch_child ) {
+      # Invoke legacy loop API
+      return $self->unwatch_child( @_ );
+   }
+
    my $childwatches = $self->{childwatches};
 
    delete $childwatches->{$pid};
@@ -2962,6 +3038,9 @@ sub unwatch_child
    }
 }
 
+# Old name
+sub unwatch_child { shift->unwatch_process( @_ ) }
+
 =head1 METHODS FOR SUBCLASSES
 
 The following methods are provided to access internal features which are
@@ -3062,17 +3141,17 @@ remaining ones in another C<extensions> parameter.
 
 For example,
 
- $loop->connect(
-    extensions => [qw( FOO BAR )],
-    %args
- )
+   $loop->connect(
+      extensions => [qw( FOO BAR )],
+      %args
+   )
 
 will become
 
- $loop->FOO_connect(
-    extensions => [qw( BAR )],
-    %args
- )
+   $loop->FOO_connect(
+      extensions => [qw( BAR )],
+      %args
+   )
 
 This is provided so that extension modules, such as L<IO::Async::SSL> can
 easily be invoked indirectly, by passing extra arguments to C<connect> methods
@@ -3082,8 +3161,8 @@ also use it.
 
 The following methods take an C<extensions> parameter:
 
- $loop->connect
- $loop->listen
+   $loop->connect
+   $loop->listen
 
 If an extension C<listen> method is invoked, it will be passed a C<listener>
 parameter even if one was not provided to the original C<< $loop->listen >>
